@@ -1,7 +1,7 @@
 /* Create a tar archive.
 
    Copyright (C) 1985, 1992, 1993, 1994, 1996, 1997, 1999, 2000, 2001,
-   2003, 2004, 2005, 2006, 2007, 2009 Free Software Foundation, Inc.
+   2003, 2004, 2005, 2006, 2007, 2009, 2010 Free Software Foundation, Inc.
 
    Written by John Gilmore, on 1985-08-25.
 
@@ -402,8 +402,7 @@ mode_to_chars (mode_t v, char *p, size_t s)
       && S_IROTH == TOREAD && S_IWOTH == TOWRITE && S_IXOTH == TOEXEC
       && archive_format != POSIX_FORMAT
       && archive_format != USTAR_FORMAT
-      && archive_format != GNU_FORMAT
-      && archive_format != OLDGNU_FORMAT)
+      && archive_format != GNU_FORMAT)
     {
       negative = v < 0;
       u = v;
@@ -516,9 +515,8 @@ write_eot (void)
 
 /* Write a "private" header */
 union block *
-start_private_header (const char *name, size_t size)
+start_private_header (const char *name, size_t size, time_t t)
 {
-  time_t t;
   union block *header = find_next_block ();
 
   memset (header->buffer, 0, sizeof (union block));
@@ -526,7 +524,6 @@ start_private_header (const char *name, size_t size)
   tar_name_copy_str (header->header.name, name, NAME_FIELD_SIZE);
   OFF_TO_CHARS (size, header->header.size);
 
-  time (&t);
   TIME_TO_CHARS (t, header->header.mtime);
   MODE_TO_CHARS (S_IFREG|S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH, header->header.mode);
   UID_TO_CHARS (getuid (), header->header.uid);
@@ -564,13 +561,13 @@ write_gnu_long_link (struct tar_stat_info *st, const char *p, char type)
   union block *header;
   char *tmpname;
 
-  header = start_private_header ("././@LongLink", size);
-  FILL(header->header.mtime, '0');
-  FILL(header->header.mode, '0');
-  FILL(header->header.uid, '0');
-  FILL(header->header.gid, '0');
-  FILL(header->header.devmajor, 0);
-  FILL(header->header.devminor, 0);
+  header = start_private_header ("././@LongLink", size, time (NULL));
+  FILL (header->header.mtime, '0');
+  FILL (header->header.mode, '0');
+  FILL (header->header.uid, '0');
+  FILL (header->header.gid, '0');
+  FILL (header->header.devmajor, 0);
+  FILL (header->header.devminor, 0);
   uid_to_uname (0, &tmpname);
   UNAME_TO_CHARS (tmpname, header->header.uname);
   free (tmpname);
@@ -605,8 +602,10 @@ split_long_name (const char *name, size_t length)
 {
   size_t i;
 
-  if (length > PREFIX_FIELD_SIZE)
+  if (length > PREFIX_FIELD_SIZE + 1)
     length = PREFIX_FIELD_SIZE + 1;
+  else if (ISSLASH (name[length - 1]))
+    length--;
   for (i = length - 1; i > 0; i--)
     if (ISSLASH (name[i]))
       break;
@@ -617,9 +616,9 @@ static union block *
 write_ustar_long_name (const char *name)
 {
   size_t length = strlen (name);
-  size_t i;
+  size_t i, nlen;
   union block *header;
-
+  
   if (length > PREFIX_FIELD_SIZE + NAME_FIELD_SIZE + 1)
     {
       ERROR ((0, 0, _("%s: file name is too long (max %d); not dumped"),
@@ -629,7 +628,7 @@ write_ustar_long_name (const char *name)
     }
 
   i = split_long_name (name, length);
-  if (i == 0 || length - i - 1 > NAME_FIELD_SIZE)
+  if (i == 0 || (nlen = length - i - 1) > NAME_FIELD_SIZE || nlen == 0)
     {
       ERROR ((0, 0,
 	      _("%s: file name is too long (cannot be split); not dumped"),
@@ -713,7 +712,8 @@ write_extended (bool global, struct tar_stat_info *st, union block *old_header)
   union block *header, hp;
   char *p;
   int type;
-
+  time_t t;
+  
   if (st->xhdr.buffer || st->xhdr.stk == NULL)
     return old_header;
 
@@ -723,13 +723,15 @@ write_extended (bool global, struct tar_stat_info *st, union block *old_header)
     {
       type = XGLTYPE;
       p = xheader_ghdr_name ();
+      time (&t);
     }
   else
     {
       type = XHDTYPE;
       p = xheader_xhdr_name (st);
+      t = st->stat.st_mtime;
     }
-  xheader_write (type, p, &st->xhdr);
+  xheader_write (type, p, t, &st->xhdr);
   free (p);
   header = find_next_block ();
   memcpy (header, &hp.buffer, sizeof (hp.buffer));
@@ -994,11 +996,9 @@ finish_header (struct tar_stat_info *st,
       && header->header.typeflag != XHDTYPE
       && header->header.typeflag != XGLTYPE)
     {
-      /* These globals are parameters to print_header, sigh.  */
-
-      current_header = header;
+      /* FIXME: This global is used in print_header, sigh.  */
       current_format = archive_format;
-      print_header (st, block_ordinal);
+      print_header (st, header, block_ordinal);
     }
 
   header = write_extended (false, st, header);
@@ -1081,7 +1081,7 @@ dump_regular_file (int fd, struct tar_stat_info *st)
 		    quotearg_colon (st->orig_file_name),
 		    STRINGIFY_BIGINT (size_left, buf)));
 	  if (! ignore_failed_read_option) 
-	    exit_status = TAREXIT_DIFFERS;
+	    set_exit_status (TAREXIT_DIFFERS);
 	  pad_archive (size_left - (bufsize - count));
 	  return dump_status_short;
 	}
@@ -1096,73 +1096,70 @@ dump_dir0 (char *directory,
 {
   dev_t our_device = st->stat.st_dev;
   const char *tag_file_name;
-  
-  if (!is_avoided_name (st->orig_file_name))
+  union block *blk = NULL;
+  off_t block_ordinal = current_block_ordinal ();
+
+  st->stat.st_size = 0;	/* force 0 size on dir */
+
+  blk = start_header (st);
+  if (!blk)
+    return;
+
+  if (incremental_option && archive_format != POSIX_FORMAT)
+    blk->header.typeflag = GNUTYPE_DUMPDIR;
+  else /* if (standard_option) */
+    blk->header.typeflag = DIRTYPE;
+
+  /* If we're gnudumping, we aren't done yet so don't close it.  */
+
+  if (!incremental_option)
+    finish_header (st, blk, block_ordinal);
+  else if (gnu_list_name->directory)
     {
-      union block *blk = NULL;
-      off_t block_ordinal = current_block_ordinal ();
-      st->stat.st_size = 0;	/* force 0 size on dir */
-
-      blk = start_header (st);
-      if (!blk)
-	return;
-
-      if (incremental_option && archive_format != POSIX_FORMAT)
-	blk->header.typeflag = GNUTYPE_DUMPDIR;
-      else /* if (standard_option) */
-	blk->header.typeflag = DIRTYPE;
-
-      /* If we're gnudumping, we aren't done yet so don't close it.  */
-
-      if (!incremental_option)
-	finish_header (st, blk, block_ordinal);
-      else if (gnu_list_name->directory)
+      if (archive_format == POSIX_FORMAT)
 	{
-	  if (archive_format == POSIX_FORMAT)
-	    {
-	      xheader_store ("GNU.dumpdir", st,
-			     safe_directory_contents (gnu_list_name->directory));
-	      finish_header (st, blk, block_ordinal);
-	    }
-	  else
-	    {
-	      off_t size_left;
-	      off_t totsize;
-	      size_t bufsize;
-	      ssize_t count;
-	      const char *buffer, *p_buffer;
-
-	      block_ordinal = current_block_ordinal ();
-	      buffer = safe_directory_contents (gnu_list_name->directory);
-	      totsize = dumpdir_size (buffer);
-	      OFF_TO_CHARS (totsize, blk->header.size);
-	      finish_header (st, blk, block_ordinal);
-	      p_buffer = buffer;
-	      size_left = totsize;
-
-	      mv_begin (st);
-	      mv_total_size (totsize);
-	      while (size_left > 0)
-		{
-		  mv_size_left (size_left);
-		  blk = find_next_block ();
-		  bufsize = available_space_after (blk);
-		  if (size_left < bufsize)
-		    {
-		      bufsize = size_left;
-		      count = bufsize % BLOCKSIZE;
-		      if (count)
-			memset (blk->buffer + size_left, 0, BLOCKSIZE - count);
-		    }
-		  memcpy (blk->buffer, p_buffer, bufsize);
-		  size_left -= bufsize;
-		  p_buffer += bufsize;
-		  set_next_block_after (blk + (bufsize - 1) / BLOCKSIZE);
-		}
-	      mv_end ();
-	    }
-	  return;
+	  xheader_store ("GNU.dumpdir", st,
+			 safe_directory_contents (gnu_list_name->directory));
+	  finish_header (st, blk, block_ordinal);
 	}
+      else
+	{
+	  off_t size_left;
+	  off_t totsize;
+	  size_t bufsize;
+	  ssize_t count;
+	  const char *buffer, *p_buffer;
+	  
+	  block_ordinal = current_block_ordinal ();
+	  buffer = safe_directory_contents (gnu_list_name->directory);
+	  totsize = dumpdir_size (buffer);
+	  OFF_TO_CHARS (totsize, blk->header.size);
+	  finish_header (st, blk, block_ordinal);
+	  p_buffer = buffer;
+	  size_left = totsize;
+	  
+	  mv_begin (st);
+	  mv_total_size (totsize);
+	  while (size_left > 0)
+	    {
+	      mv_size_left (size_left);
+	      blk = find_next_block ();
+	      bufsize = available_space_after (blk);
+	      if (size_left < bufsize)
+		{
+		  bufsize = size_left;
+		  count = bufsize % BLOCKSIZE;
+		  if (count)
+		    memset (blk->buffer + size_left, 0, BLOCKSIZE - count);
+		}
+	      memcpy (blk->buffer, p_buffer, bufsize);
+	      size_left -= bufsize;
+	      p_buffer += bufsize;
+	      set_next_block_after (blk + (bufsize - 1) / BLOCKSIZE);
+	    }
+	  mv_end ();
+	}
+      return;
     }
 
   if (!recursion_option)
@@ -1334,7 +1331,7 @@ create_archive (void)
 
   write_eot ();
   close_archive ();
-
+  finish_deferred_unlinks ();
   if (listed_incremental_option)
     write_directory_file ();
 }
@@ -1365,7 +1362,7 @@ unknown_file_error (char const *p)
 	   (0, 0, _("%s: Unknown file type; file ignored"),
 	    quotearg_colon (p)));
   if (!ignore_failed_read_option)
-    exit_status = TAREXIT_FAILURE;
+    set_exit_status (TAREXIT_FAILURE);
 }
 
 
@@ -1414,8 +1411,8 @@ dump_hard_link (struct tar_stat_info *st)
 	  blk->header.typeflag = LNKTYPE;
 	  finish_header (st, blk, block_ordinal);
 
-	  if (remove_files_option && unlink (st->orig_file_name) != 0)
-	    unlink_error (st->orig_file_name);
+	  if (remove_files_option)
+	    queue_deferred_unlink (st->orig_file_name, false);
 
 	  return true;
 	}
@@ -1557,9 +1554,6 @@ dump_file0 (struct tar_stat_info *st, const char *p,
       return;
     }
 
-  if (is_avoided_name (p))
-    return;
-
   is_dir = S_ISDIR (st->stat.st_mode) != 0;
 
   if (!is_dir && dump_hard_link (st))
@@ -1667,8 +1661,7 @@ dump_file0 (struct tar_stat_info *st, const char *p,
 	      WARNOPT (WARN_FILE_CHANGED,
 		       (0, 0, _("%s: file changed as we read it"),
 			quotearg_colon (p)));
-	      if (exit_status == TAREXIT_SUCCESS)
-		exit_status = TAREXIT_DIFFERS;
+	      set_exit_status (TAREXIT_DIFFERS);
 	    }
 	  else if (atime_preserve_option == replace_atime_preserve
 		   && set_file_atime (fd, p, restore_times) != 0)
@@ -1682,18 +1675,7 @@ dump_file0 (struct tar_stat_info *st, const char *p,
 	}
 
       if (ok && remove_files_option)
-	{
-	  if (is_dir)
-	    {
-	      if (rmdir (p) != 0 && errno != ENOTEMPTY)
-		rmdir_error (p);
-	    }
-	  else
-	    {
-	      if (unlink (p) != 0)
-		unlink_error (p);
-	    }
-	}
+	queue_deferred_unlink (p, is_dir);
 
       return;
     }
@@ -1729,10 +1711,8 @@ dump_file0 (struct tar_stat_info *st, const char *p,
       /* nothing more to do to it */
 
       if (remove_files_option)
-	{
-	  if (unlink (p) == -1)
-	    unlink_error (p);
-	}
+	queue_deferred_unlink (p, false);
+
       file_count_links (st);
       return;
     }
@@ -1784,10 +1764,7 @@ dump_file0 (struct tar_stat_info *st, const char *p,
 
   finish_header (st, header, block_ordinal);
   if (remove_files_option)
-    {
-      if (unlink (p) == -1)
-	unlink_error (p);
-    }
+    queue_deferred_unlink (p, false);
 }
 
 void
